
class EmptyValueError(ValueError): pass

class UnexpectedValueError(ValueError): pass

class InvalidConfig(Exception): pass

class HttpException(Exception): pass

class HttpHeaderTooLarge(HttpException): pass

class HttpPayloadTooLarge(HttpException): pass

class HttpNotFound(HttpException): pass

class HttpMethodNotAllowed(HttpException): pass

class NetworkError(RuntimeError): pass



import logging

from cachetools import TTLCache
from dnslib import QTYPE

from . import Middleware


logger = logging.getLogger(__name__)


class CacheMiddleware(Middleware):
    def __init__(self, next, max_size, ttl):
        super().__init__(next)
        self.cache = TTLCache(maxsize=max_size, ttl=ttl)

    async def handle(self, request, **kwargs):
        cache_key = '%s;%s' % (request.q.qname, QTYPE[request.q.qtype])
        if cache_key in self.cache:
            logger.debug('Cache hit "%s"', cache_key)
            response = self.cache[cache_key]
            response.header.id = request.header.id
            return response

        logger.debug('Cache miss "%s"', cache_key)
        response = await super().handle(request, **kwargs)
        if response is not None:
            self.cache[cache_key] = response
        return response



import io
import sys
import asyncio
import pathlib


__all__ = ('Middleware', 'CacheMiddleware', 'HostsMiddleware', 'BlockMiddleware', 'RulesMiddleware')


def _middleware_name_to_class_name(name):
    return ''.join(
        [_.capitalize() for _ in name.split('_')]
    ) + 'Middleware'


def create_middleware(name, next, config):
    class_name = _middleware_name_to_class_name(name)
    if class_name not in __all__: return next
    middleware_class = getattr(sys.modules[__name__], class_name)
    if config is None:
        return middleware_class(next)
    elif isinstance(config, list):
        return middleware_class(next, *config)
    elif isinstance(config, dict):
        return middleware_class(next, **config)
    return middleware_class(next, config)


async def load_config(filename, server, parser):
    if filename.startswith(('http://', 'https://')):
        response = await server.fetch(filename)
        fp = io.StringIO(response.body.decode())
    else:
        fp = pathlib.Path(filename).open('r')
    with fp:
        return await asyncio.to_thread(parser, fp)


class Middleware:
    @property
    def server(self):
        if hasattr(self, '_server'):
            return self._server
        from ..server import Server
        if isinstance(self.next, Server):
            self._server = self.next
            return self._server
        self._server =  self.next.server
        return self._server

    def get_component(self, name):
        class_name = _middleware_name_to_class_name(name)
        if __class__.__name__ == class_name:
            return self
        if not hasattr(self.next, self.get_component.__name__):
            raise ValueError('Middleware %s not found' % (name, ))
        return self.next.get_component(name)

    def __init__(self, next):
        self.next = next

    def abort(self):
        return self.next.abort()

    def abort(self):
        return self.next.abort()

    async def handle(self, request, **kwargs):
        return await self.next.handle(request, **kwargs)

    async def bootstrap(self):
        return await self.next.bootstrap()

    async def restart(self):
        return await self.next.restart()

    async def reload(self):
        return await self.next.reload()


from .cache import CacheMiddleware
from .hosts import HostsMiddleware
from .block import BlockMiddleware
from .rules import RulesMiddleware



import asyncio
import logging
import re
import enum
import ipaddress
import functools
import subprocess

from collections.abc import Iterable

import dnslib
import pyparsing

from . import Middleware
from ..exceptions import InvalidConfig
from ..helpers import split_domain


logger = logging.getLogger(__name__)


opt_comma = pyparsing.Suppress(pyparsing.Opt(','))

_dec_num = pyparsing.Word(pyparsing.nums, max=3)

_ipv4_literal = _dec_num - ('.' + _dec_num) * 3

_ipv6_literal = pyparsing.Word(pyparsing.hexnums + ':.')

_ipv4_cidr_literal = _ipv4_literal + '/' + pyparsing.Word(pyparsing.nums, max=2)

_ipv6_cidr_literal = _ipv6_literal + '/' + pyparsing.Word(pyparsing.nums, max=3)

ip_literal = pyparsing.Combine(_ipv4_literal | _ipv6_literal)\
    .set_name('ip address')\
    .set_parse_action(lambda tokens: ipaddress.ip_address(tokens[0]))

ip_cidr_literal = pyparsing.Combine(_ipv4_cidr_literal | _ipv6_cidr_literal)\
    .set_name('CIDR')\
    .set_parse_action(lambda tokens: ipaddress.ip_network(tokens[0], False))

ip_cidr_literal_list = (pyparsing.Suppress('(') + pyparsing.OneOrMore(ip_cidr_literal + opt_comma) + pyparsing.Suppress(')'))\
    .set_name('list of CIDRs')\
    .set_parse_action(lambda tokens: tuple(tokens))

country_code_literal = pyparsing.Word(pyparsing.alphas, exact=2).set_name('country code')

domain_literal = pyparsing.Word(pyparsing.alphanums + '.-').set_name('domain')\
    .set_parse_action(lambda tokens: tokens[0].lower())

domain_literal_set = (pyparsing.Suppress('(') + pyparsing.OneOrMore(domain_literal + opt_comma) + pyparsing.Suppress(')'))\
    .set_name('set of domains')\
    .set_parse_action(lambda tokens: set(map(lambda _: _.strip('.'), tokens)))

domain_literal_list = (pyparsing.Suppress('(') + pyparsing.OneOrMore(domain_literal + opt_comma) + pyparsing.Suppress(')'))\
    .set_name('list of domains')\
    .set_parse_action(lambda tokens: tuple(tokens))

ip_literal_list = (pyparsing.Suppress('(') + pyparsing.OneOrMore(ip_literal + opt_comma) + pyparsing.Suppress(')'))\
    .set_name('list of ip addresses')\
    .set_parse_action(lambda tokens: tuple(tokens))

ip_literal_set = (pyparsing.Suppress('(') + pyparsing.OneOrMore(ip_literal + opt_comma) + pyparsing.Suppress(')'))\
    .set_name('set of ip addresses')\
    .set_parse_action(lambda tokens: set(tokens))

always_literal = pyparsing.Literal('always').set_name('always')\
    .set_parse_action(lambda _: True)


def _and_op(op, tokens):
    return op, tuple(filter(lambda it: it != 'and', tokens))


def _or_op(op, tokens):
    return op, tuple(filter(lambda it: it != 'or', tokens))


def _response_add_answers(response, ip_addrs):
    hostname = response.q.qname.idna().rstrip('.')
    qtype = response.q.qtype
    if not isinstance(ip_addrs, Iterable):
        ip_addrs = (ip_addrs, )
    for ip in ip_addrs:
        if isinstance(ip, ipaddress.IPv6Address) and qtype == dnslib.QTYPE.AAAA:
            response.add_answer(dnslib.RR(
                hostname, qtype, 
                rdata=dnslib.AAAA(str(ip))
            ))
        elif isinstance(ip, ipaddress.IPv4Address) and qtype == dnslib.QTYPE.A:
            response.add_answer(dnslib.RR(
                hostname, qtype, 
                rdata=dnslib.A(str(ip))
            ))
    return response


def _response_nxdomain(response):
    response.rr = []
    response.set_header_qa()
    response.header.rcode = dnslib.RCODE.NXDOMAIN
    return response


class OpCode(enum.Enum):
    NOT = enum.auto()
    AND = enum.auto()
    OR = enum.auto()
    DOMAIN_CONTAINS = enum.auto()
    DOMAIN_NOT_CONTAINS = enum.auto()
    DOMAIN_EQ = enum.auto()
    DOMAIN_NOT_EQ = enum.auto()
    DOMAIN_STARTS_WITH = enum.auto()
    DOMAIN_STARTS_WITHOUT = enum.auto()
    DOMAIN_ENDS_WITH = enum.auto()
    DOMAIN_ENDS_WITHOUT = enum.auto()
    DOMAIN_MATCH = enum.auto()
    BLOCK = enum.auto()
    RETURN = enum.auto()
    UPSTREAM_GROUP_SET = enum.auto()
    UPSTREAM_NAME_SET = enum.auto()
    DOMAIN_REPLACE = enum.auto()
    FIRST = enum.auto()
    LAST = enum.auto()
    IP_IN = enum.auto()
    IP_NOT_IN = enum.auto()
    IP_EQ = enum.auto()
    IP_NOT_EQ = enum.auto()
    GEOIP_EQ = enum.auto()
    GEOIP_NOT_EQ = enum.auto()
    ANY_IP_IN = enum.auto()
    ANY_IP_NOT_IN = enum.auto()
    ANY_IP_EQ = enum.auto()
    ANY_IP_NOT_EQ = enum.auto()
    ANY_GEOIP_EQ = enum.auto()
    ANY_GEOIP_NOT_EQ = enum.auto()
    ALL_IP_IN = enum.auto()
    ALL_IP_NOT_IN = enum.auto()
    ALL_IP_EQ = enum.auto()
    ALL_IP_NOT_EQ = enum.auto()
    ALL_GEOIP_EQ = enum.auto()
    ALL_GEOIP_NOT_EQ = enum.auto()
    RECORD_ADD = enum.auto()
    RECORD_ADD_IF = enum.auto()
    RECORD_REMOVE_WHERE = enum.auto()
    RECORD_REPLACE_WHERE = enum.auto()
    RUN_WHERE = enum.auto()
    BLOCK_IF = enum.auto()
    RETURN_IF = enum.auto()

    def __repr__(self):
        return self.name


class RuleIfParser:
    class IfExpr:
        def __init__(self, ast):
            self.ast = ast

        def __repr__(self):
            return repr(self.ast)

        def test(self, domain):
            def _test(ast):
                if not isinstance(ast, tuple):
                    return ast
                opcode = ast[0]
                if opcode == OpCode.AND:
                    return all(map(_test, ast[1]))
                elif opcode == OpCode.OR:
                    return any(map(_test, ast[1]))
                elif opcode == OpCode.NOT:
                    return not _test(ast[1])
                elif opcode == OpCode.DOMAIN_CONTAINS:
                    if isinstance(ast[1], tuple):
                        return any(map(lambda _: _ in domain, ast[1]))
                    else:
                        return ast[1] in domain
                elif opcode == OpCode.DOMAIN_NOT_CONTAINS:
                    if isinstance(ast[1], tuple):
                        return all(map(lambda _: _ not in domain, ast[1]))
                    else:
                        return ast[1] not in domain
                elif opcode == OpCode.DOMAIN_EQ:
                    if isinstance(ast[1], set):
                        return domain in ast[1]
                    else:
                        return ast[1] == domain
                elif opcode == OpCode.DOMAIN_NOT_EQ:
                    if isinstance(ast[1], set):
                        return domain not in ast[1]
                    else:
                        return ast[1] != domain
                elif opcode == OpCode.DOMAIN_STARTS_WITH:
                    if isinstance(ast[1], set):
                        return any(map(lambda _: _ in ast[1], split_domain(domain, False)))
                    else:
                        return domain.startswith(ast[1])
                elif opcode == OpCode.DOMAIN_STARTS_WITHOUT:
                    if isinstance(ast[1], set):
                        return all(map(lambda _: _ not in ast[1], split_domain(domain, False)))
                    else:
                        return not domain.startswith(ast[1])
                elif opcode == OpCode.DOMAIN_ENDS_WITH:
                    if isinstance(ast[1], set):
                        return any(map(lambda _: _ in ast[1], split_domain(domain)))
                    else:
                        return domain.endswith(ast[1])
                elif opcode == OpCode.DOMAIN_ENDS_WITHOUT:
                    if isinstance(ast[1], set):
                        return all(map(lambda _: _ not in ast[1], split_domain(domain)))
                    else:
                        return not domain.endswith(ast[1])
                elif opcode == OpCode.DOMAIN_MATCH:
                    return re.fullmatch(ast[1], domain) is not None
                return False
            return _test(self.ast)

    def __init__(self):
        expr_domain_contains = ((domain_literal | domain_literal_list) + 'in domain')\
            .set_name('domain contains expression')\
            .set_parse_action(lambda tokens: (OpCode.DOMAIN_CONTAINS, tokens[0]))
        expr_domain_not_contains = ((domain_literal | domain_literal_list) + 'not in domain')\
            .set_name('domain not contains expression')\
            .set_parse_action(lambda tokens: (OpCode.DOMAIN_NOT_CONTAINS, tokens[0]))
        expr_domain_equal = ('domain is' + (domain_literal | domain_literal_set))\
            .set_name('domain equal expression')\
            .set_parse_action(lambda tokens: (OpCode.DOMAIN_EQ, tokens[1]))
        expr_domain_not_equal = ('domain is not' + (domain_literal | domain_literal_set))\
            .set_name('domain not equal expression')\
            .set_parse_action(lambda tokens: (OpCode.DOMAIN_NOT_EQ, tokens[1]))
        expr_domain_starts_with = ('domain starts with' + (domain_literal | domain_literal_set))\
            .set_name('domain starts with expression')\
            .set_parse_action(lambda tokens: (OpCode.DOMAIN_STARTS_WITH, tokens[1]))
        expr_domain_starts_without = ('domain starts without' + (domain_literal | domain_literal_set))\
            .set_name('domain starts without expression')\
            .set_parse_action(lambda tokens: (OpCode.DOMAIN_STARTS_WITHOUT, tokens[1]))
        expr_domain_ends_with = ('domain ends with' + (domain_literal | domain_literal_set))\
            .set_name('domain ends with expression')\
            .set_parse_action(lambda tokens: (OpCode.DOMAIN_ENDS_WITH, tokens[1]))
        expr_domain_ends_without = ('domain ends without' + (domain_literal | domain_literal_set))\
            .set_name('domain ends without expression')\
            .set_parse_action(lambda tokens: (OpCode.DOMAIN_ENDS_WITHOUT, tokens[1]))
        expr_domain_match = ('domain match' + pyparsing.QuotedString('/'))\
            .set_name('domain match expression')\
            .set_parse_action(lambda tokens: (OpCode.DOMAIN_MATCH, tokens[1]))
        expr_simple = (
            expr_domain_not_contains |
            expr_domain_contains |
            expr_domain_not_equal |
            expr_domain_equal |
            expr_domain_starts_without |
            expr_domain_starts_with |
            expr_domain_ends_without |
            expr_domain_ends_with |
            expr_domain_match
        )
        expr = pyparsing.Forward()
        expr <<= pyparsing.infix_notation(expr_simple, [
            ('not', 1, pyparsing.opAssoc.RIGHT, lambda tokens: (OpCode.NOT, tokens[0][1])),
            ('and', 2, pyparsing.opAssoc.LEFT, lambda tokens: _and_op(OpCode.AND, tokens[0])),
            ('or', 2, pyparsing.opAssoc.LEFT, lambda tokens: _or_op(OpCode.OR, tokens[0])),
        ])
        self.parser = expr | always_literal
    
    def parse(self, source):
        parse_results = self.parser.parse_string(source, parse_all=True)
        return self.IfExpr(parse_results[0])


class RuleThenParser:
    class ThenExec:
        def __repr__(self):
            return repr(self.ast)

        def __init__(self, ast):
            self.ast = ast[0]

        def exec(self, request):
            response = request.reply()
            _exec = self.ast[0]
            if _exec == OpCode.BLOCK:
                response.header.rcode = dnslib.RCODE.NXDOMAIN
            elif _exec == OpCode.RETURN:
                ip_addrs = self.ast[1] if isinstance(self.ast[1], tuple) else (self.ast[1], )
                _response_add_answers(response, ip_addrs)
            return response

    def __init__(self):
        stm_block = pyparsing.Literal('block')\
            .set_name('block statement')\
            .set_parse_action(lambda _: (OpCode.BLOCK, ))
        stm_return = ('return' + (ip_literal_list | ip_literal))\
            .set_name('return statement')\
            .set_parse_action(lambda tokens: (OpCode.RETURN, tokens[1]))
        stm_single = (
            stm_block |
            stm_return
        )
        self.parser = stm_single

    def parse(self, source):
        parse_results = self.parser.parse_string(source, parse_all=True)
        return self.ThenExec(parse_results)


class AfterHandler:
    def after_handle(self, response, geoip, ipaddr, run):
        return response


class ReplacDomainHandler(AfterHandler):
    def __init__(self, domain):
        self.domain = domain

    def after_handle(self, response, geoip, ipaddr, run):
        response.q.qname = self.domain
        return response


class AfterHandlerList:
    def __init__(self):
        self.handlers = []

    def join(self, handler):
        if handler is None or not hasattr(handler, 'after_handle'):
            return self
        self.handlers.append(handler)
        return self

    def after_handle(self, response, geoip, ipaddr, run):
        for handler in self.handlers:
            response = handler.after_handle(response, geoip, ipaddr, run)
        return response


class RuleBeforeParser:
    class BeforeExec:
        def __repr__(self):
            return repr(self.ast)

        def __init__(self, ast):
            self.ast = list(ast)

        def exec(self, request, kwargs):
            after_handler = None
            for stm in self.ast:
                _exec = stm[0]
                if _exec == OpCode.UPSTREAM_GROUP_SET:
                    kwargs['upstream_group'] = stm[1]
                elif _exec == OpCode.UPSTREAM_NAME_SET:
                    kwargs['upstream_name'] = stm[1]
                elif _exec == OpCode.DOMAIN_REPLACE:
                    after_handler = ReplacDomainHandler(request.q.qname)
                    request.q.qname = stm[1]
            return after_handler

    def __init__(self):
        upstream_name_literal = pyparsing.Word(pyparsing.alphanums + '.-_')\
            .set_name('upstream name')
        upstream_group_literal = pyparsing.Word(pyparsing.alphanums + '.-_')\
            .set_name('upstream group')
        stm_upstream_name_set = ('set upstream name to' + upstream_name_literal)\
            .set_name('set upstream name statement')\
            .set_parse_action(lambda tokens: (OpCode.UPSTREAM_NAME_SET, tokens[1]))
        stm_upstream_group_set = ('set upstream group to' + upstream_group_literal)\
            .set_name('set upstream group statement')\
            .set_parse_action(lambda tokens: (OpCode.UPSTREAM_GROUP_SET, tokens[1]))
        stm_domain_replace = ('replace domain by' + domain_literal)\
            .set_name('replace domain statement')\
            .set_parse_action(lambda tokens: (OpCode.DOMAIN_REPLACE, tokens[1]))
        stm_simple = (
            stm_upstream_name_set |
            stm_upstream_group_set |
            stm_domain_replace
        )
        self.parser = pyparsing.OneOrMore(stm_simple + opt_comma)

    def parse(self, source):
        parse_results = self.parser.parse_string(source, parse_all=True)
        return self.BeforeExec(parse_results)


class RuleAfterParser:
    class AfterExec:
        def __init__(self, ast):
            self.ast = list(ast)

        def __repr__(self):
            return repr(self.ast)

        def test_record(self, geoip, ipaddr, response, expr_if, idx_rr):
            idx, rr = idx_rr
            def test(ast):
                if not isinstance(ast, tuple):
                    return ast
                if rr.rtype not in (dnslib.QTYPE.AAAA, dnslib.QTYPE.A):
                    return False

                opcode = ast[0]
                if opcode == OpCode.AND:
                    return all(map(test, ast[1]))
                elif opcode == OpCode.OR:
                    return any(map(test, ast[1]))
                elif opcode == OpCode.NOT:
                    return not test(ast[1])
                elif opcode == OpCode.IP_EQ:
                    if isinstance(ast[1], set):
                        return ipaddr(rr) in ast[1]
                    else:
                        return ast[1] == ipaddr(rr)
                elif opcode == OpCode.IP_NOT_EQ:
                    if isinstance(ast[1], set):
                        return ipaddr(rr) not in ast[1]
                    else:
                        return ast[1] != ipaddr(rr)
                elif opcode == OpCode.IP_IN:
                    if isinstance(ast[1], tuple):
                        return any(map(lambda _: ipaddr(rr) in _, ast[1]))
                    else:
                        return ipaddr(rr) in ast[1]
                elif opcode == OpCode.IP_NOT_IN:
                    if isinstance(ast[1], tuple):
                        return all(map(lambda _: ipaddr(rr) not in _, ast[1]))
                    else:
                        return ipaddr(rr) not in ast[1]
                elif opcode == OpCode.GEOIP_EQ:
                    return geoip(rr) == ast[1].upper()
                elif opcode == OpCode.GEOIP_NOT_EQ:
                    return geoip(rr) != ast[1].upper()
                elif opcode == OpCode.FIRST:
                    return idx == 0
                elif opcode == OpCode.LAST:
                    return idx == len(response.rr) - 1
                return False
            return test(expr_if)

        def test_response(self, geoip, ipaddr, expr_if, response):
            def test(ast):
                if not isinstance(ast, tuple):
                    return ast

                opcode = ast[0]
                if opcode == OpCode.AND:
                    return all(map(test, ast[1]))
                elif opcode == OpCode.OR:
                    return any(map(test, ast[1]))
                elif opcode == OpCode.NOT:
                    return not test(ast[1])

                rr = filter(lambda rr: rr.rtype in (dnslib.QTYPE.AAAA, dnslib.QTYPE.A), response.rr)
                ip_addresses = map(lambda rr: ipaddr(rr), rr)
                country_codes = map(lambda rr: geoip(rr), rr)
                if opcode == OpCode.ANY_IP_EQ:
                    if isinstance(ast[1], set):
                        return any(map(lambda _: _ in ast[1], ip_addresses))
                    else:
                        return any(map(lambda ip: ip == ast[1], ip_addresses))
                elif opcode == OpCode.ANY_IP_NOT_EQ:
                    if isinstance(ast[1], set):
                        return any(map(lambda _: _ not in ast[1], ip_addresses))
                    else:
                        return any(map(lambda ip: ip != ast[1], ip_addresses))
                elif opcode == OpCode.ANY_IP_IN:
                    if isinstance(ast[1], tuple):
                        return any(map(lambda ip: any(map(lambda _: ip in _, ast[1])), ip_addresses))
                    else:
                        return any(map(lambda ip: ip in ast[1], ip_addresses))
                elif opcode == OpCode.ANY_IP_NOT_IN:
                    if isinstance(ast[1], tuple):
                        return any(map(lambda ip: all(map(lambda _: ip not in _, ast[1])), ip_addresses))
                    else:
                        return any(map(lambda ip: ip not in ast[1], ip_addresses))
                elif opcode == OpCode.ANY_GEOIP_EQ:
                    return any(map(lambda country_code: country_code == ast[1], country_codes))
                elif opcode == OpCode.ANY_GEOIP_NOT_EQ:
                    return any(map(lambda country_code: country_code != ast[1], country_codes))
                elif opcode == OpCode.ALL_IP_EQ:
                    if isinstance(ast[1], set):
                        return all(map(lambda _: _ in ast[1], ip_addresses))
                    else:
                        return all(map(lambda ip: ip == ast[1], ip_addresses))
                elif opcode == OpCode.ALL_IP_NOT_EQ:
                    if isinstance(ast[1], set):
                        return all(map(lambda _: _ not in ast[1], ip_addresses))
                    else:
                        return all(map(lambda ip: ip != ast[1], ip_addresses))
                elif opcode == OpCode.ALL_IP_IN:
                    if isinstance(ast[1], tuple):
                        return all(map(lambda ip: any(map(lambda _: ip in _, ast[1])), ip_addresses))
                    else:
                        return all(map(lambda ip: ip in ast[1], ip_addresses))
                elif opcode == OpCode.ALL_IP_NOT_IN:
                    if isinstance(ast[1], tuple):
                        return all(map(lambda ip: all(map(lambda _: ip not in _, ast[1])), ip_addresses))
                    else:
                        return all(map(lambda ip: ip not in ast[1], ip_addresses))
                elif opcode == OpCode.ALL_GEOIP_EQ:
                    return all(map(lambda country_code: country_code == ast[1], country_codes))
                elif opcode == OpCode.ALL_GEOIP_NOT_EQ:
                    return all(map(lambda country_code: country_code != ast[1], country_codes))
                return False
            return test(expr_if)

        def exec(self, response, geoip, ipaddr, run):
            test_record = functools.partial(self.test_record, geoip, ipaddr, response)
            test_response = functools.partial(self.test_response, geoip, ipaddr)

            for stm in self.ast:
                _exec = stm[0]
                if _exec == OpCode.RECORD_ADD:
                    _response_add_answers(response, stm[1])
                    continue

                expr_if = stm[1]
                if _exec == OpCode.RECORD_REMOVE_WHERE:
                    response.rr = list(map(
                        lambda idx_rr: idx_rr[1],
                        filter(
                            lambda idx_rr: not test_record(expr_if, idx_rr), 
                            enumerate(response.rr)
                        )
                    ))
                    response.set_header_qa()
                    continue
                elif _exec == OpCode.RECORD_REPLACE_WHERE:
                    qtype = response.q.qtype
                    ip = stm[2]
                    for idx, rr in enumerate(response.rr):
                        if not test_record(expr_if, (idx, rr)):
                            continue
                        if isinstance(ip, ipaddress.IPv6Address) and qtype == dnslib.QTYPE.AAAA:
                            rr.rdata = dnslib.AAAA(str(ip))
                        elif isinstance(ip, ipaddress.IPv4Address) and qtype == dnslib.QTYPE.A:
                            rr.rdata = dnslib.A(str(ip))
                    continue
                elif _exec == OpCode.RUN_WHERE:
                    cmd = stm[2]
                    hostname = response.q.qname.idna().rstrip('.')
                    for idx, rr in enumerate(response.rr):
                        if not test_record(expr_if, (idx, rr)):
                            continue
                        run(cmd.format(ip=rr.rdata, domain=hostname))
                    continue

                if _exec == OpCode.RETURN_IF:
                    if test_response(expr_if, response):
                        response.rr = []
                        _response_add_answers(response, stm[2])
                elif _exec == OpCode.BLOCK_IF:
                    if test_response(expr_if, response):
                        _response_nxdomain(response)
                elif _exec == OpCode.RECORD_ADD_IF:
                    if test_response(expr_if, response):
                        _response_add_answers(response, stm[2])
            return response

    def __init__(self):
        expr_first = pyparsing.Literal('first').set_name('first')\
            .set_parse_action(lambda _: (OpCode.FIRST, ))
        expr_last = pyparsing.Literal('last').set_name('last')\
            .set_parse_action(lambda _: (OpCode.LAST, ))
        expr_ip_in = ('ip in' + (ip_cidr_literal | ip_cidr_literal_list))\
            .set_name('ip in cidr expression')\
            .set_parse_action(lambda tokens: (OpCode.IP_IN, tokens[1]))
        expr_ip_not_in = ('ip not in' + (ip_cidr_literal | ip_cidr_literal_list))\
            .set_name('ip not in cidr expression')\
            .set_parse_action(lambda tokens: (OpCode.IP_NOT_IN, tokens[1]))
        expr_ip_equal = ('ip is' + (ip_literal | ip_literal_set))\
            .set_name('ip equal expression')\
            .set_parse_action(lambda tokens: (OpCode.IP_EQ, tokens[1]))
        expr_ip_not_equal = ('ip is not' + (ip_literal | ip_literal_set))\
            .set_name('ip not equal expression')\
            .set_parse_action(lambda tokens: (OpCode.IP_NOT_EQ, tokens[1]))
        expr_geoip_equal = ('geoip is' + country_code_literal)\
            .set_name('geoip equal expression')\
            .set_parse_action(lambda tokens: (OpCode.GEOIP_EQ, tokens[1]))
        expr_geoip_not_equal = ('geoip is not' + country_code_literal)\
            .set_name('geoip not equal expression')\
            .set_parse_action(lambda tokens: (OpCode.GEOIP_NOT_EQ, tokens[1]))
        expr_any_ip_in = ('any ip in' + (ip_cidr_literal | ip_cidr_literal_list))\
            .set_name('any ip in cidr expression')\
            .set_parse_action(lambda tokens: (OpCode.ANY_IP_IN, tokens[1]))
        expr_any_ip_not_in = ('any ip not in' + (ip_cidr_literal | ip_cidr_literal_list))\
            .set_name('any ip not in cidr expression')\
            .set_parse_action(lambda tokens: (OpCode.ANY_IP_NOT_IN, tokens[1]))
        expr_any_ip_equal = ('any ip is' + (ip_literal | ip_literal_set))\
            .set_name('any ip equal expression')\
            .set_parse_action(lambda tokens: (OpCode.ANY_IP_EQ, tokens[1]))
        expr_any_ip_not_equal = ('any ip is not' + (ip_literal | ip_literal_set))\
            .set_name('any ip not equal expression')\
            .set_parse_action(lambda tokens: (OpCode.ANY_IP_NOT_EQ, tokens[1]))
        expr_any_geoip_equal = ('any geoip is' + country_code_literal)\
            .set_name('any geoip equal expression')\
            .set_parse_action(lambda tokens: (OpCode.ANY_GEOIP_EQ, tokens[1]))
        expr_any_geoip_not_equal = ('any geoip is not' + country_code_literal)\
            .set_name('any geoip not equal expression')\
            .set_parse_action(lambda tokens: (OpCode.ANY_GEOIP_NOT_EQ, tokens[1]))
        expr_all_ip_in = ('all ip in' + (ip_cidr_literal | ip_cidr_literal_list))\
            .set_name('all ip in cidr expression')\
            .set_parse_action(lambda tokens: (OpCode.ALL_IP_IN, tokens[1]))
        expr_all_ip_not_in = ('all ip not in' + (ip_cidr_literal | ip_cidr_literal_list))\
            .set_name('all ip not in cidr expression')\
            .set_parse_action(lambda tokens: (OpCode.ALL_IP_NOT_IN, tokens[1]))
        expr_all_ip_equal = ('all ip is' + (ip_literal | ip_literal_set))\
            .set_name('all ip equal expression')\
            .set_parse_action(lambda tokens: (OpCode.ALL_IP_EQ, tokens[1]))
        expr_all_ip_not_equal = ('all ip is not' + (ip_literal | ip_literal_set))\
            .set_name('all ip not equal expression')\
            .set_parse_action(lambda tokens: (OpCode.ALL_IP_NOT_EQ, tokens[1]))
        expr_all_geoip_equal = ('all geoip is' + country_code_literal)\
            .set_name('all geoip equal expression')\
            .set_parse_action(lambda tokens: (OpCode.ALL_GEOIP_EQ, tokens[1]))
        expr_all_geoip_not_equal = ('all geoip is not' + country_code_literal)\
            .set_name('all geoip not equal expression')\
            .set_parse_action(lambda tokens: (OpCode.ALL_GEOIP_NOT_EQ, tokens[1]))
        expr_simple = (
            expr_first |
            expr_last |
            expr_ip_not_in |
            expr_ip_in |
            expr_ip_not_equal |
            expr_ip_equal |
            expr_geoip_not_equal |
            expr_geoip_equal
        )
        expr_complex = (
            expr_any_ip_in |
            expr_any_ip_not_in |
            expr_any_ip_equal |
            expr_any_ip_not_equal |
            expr_any_geoip_not_equal |
            expr_any_geoip_equal |
            expr_all_ip_in |
            expr_all_ip_not_in |
            expr_all_ip_equal |
            expr_all_ip_not_equal |
            expr_all_geoip_not_equal |
            expr_all_geoip_equal
        )
        expr_if = pyparsing.Forward()
        expr_if <<= (pyparsing.infix_notation(expr_complex, [
            ('not', 1, pyparsing.opAssoc.RIGHT, lambda tokens: (OpCode.NOT, tokens[0][1])),
            ('and', 2, pyparsing.opAssoc.LEFT, lambda tokens: _and_op(OpCode.AND, tokens[0])),
            ('or', 2, pyparsing.opAssoc.LEFT, lambda tokens: _or_op(OpCode.OR, tokens[0])),
        ]) | always_literal)
        expr_where = pyparsing.Forward()
        expr_where <<= (pyparsing.infix_notation(expr_simple, [
            ('not', 1, pyparsing.opAssoc.RIGHT, lambda tokens: (OpCode.NOT, tokens[0][1])),
            ('and', 2, pyparsing.opAssoc.LEFT, lambda tokens: _and_op(OpCode.AND, tokens[0])),
            ('or', 2, pyparsing.opAssoc.LEFT, lambda tokens: _or_op(OpCode.OR, tokens[0])),
        ]) | always_literal)
        stm_record_add = ('add record' + (ip_literal | ip_literal_list))\
            .set_name('add record statement')\
            .set_parse_action(lambda tokens: (OpCode.RECORD_ADD, tokens[1]))
        stm_record_add_if = ('add record' + (ip_literal | ip_literal_list) + 'if' + expr_if)\
            .set_name('add record if statement')\
            .set_parse_action(lambda tokens: (OpCode.RECORD_ADD_IF, tokens[3], tokens[1]))
        stm_record_remove_where = ('remove record where' + expr_where)\
            .set_name('remove record where statement')\
            .set_parse_action(lambda tokens: (OpCode.RECORD_REMOVE_WHERE, *tokens[1:]))
        stm_record_replace_where = ('replace record by' + ip_literal + 'where' + expr_where)\
            .set_name('replace record where statement')\
            .set_parse_action(lambda tokens: (OpCode.RECORD_REPLACE_WHERE, tokens[3], tokens[1]))
        stm_run_command_where = ('run' + pyparsing.QuotedString('"') + 'where' + expr_where)\
            .set_name('run command where statement')\
            .set_parse_action(lambda tokens: (OpCode.RUN_WHERE, tokens[3], tokens[1]))
        stm_simple = (
            stm_record_add_if |
            stm_record_add |
            stm_record_remove_where |
            stm_record_replace_where |
            stm_run_command_where
        )
        stm_block_if = ('block if' + expr_if)\
            .set_name('block if statement')\
            .set_parse_action(lambda tokens: (OpCode.BLOCK_IF, tokens[1]))
        stm_return_if = ('return' + (ip_literal_list | ip_literal) + 'if' + expr_if)\
            .set_name('return if statement')\
            .set_parse_action(lambda tokens: (OpCode.RETURN_IF, tokens[3], tokens[1]))
        stm_single = (
            stm_block_if |
            stm_return_if
        )
        self.parser = stm_single | pyparsing.OneOrMore(stm_simple + opt_comma)

    def parse(self, source):
        parse_results = self.parser.parse_string(source, parse_all=True)
        return self.AfterExec(parse_results)


class RulesMiddleware(Middleware):
    class Rule:
        def test(self, domain):
            return self.if_expr.test(domain)

    class IfThenRule(Rule):
        def __init__(self, if_expr, then_exec, is_ended):
            self.if_expr = if_expr
            self.then_exec = then_exec
            self.is_ended = is_ended

        def exec(self, request):
            return self.then_exec.exec(request)

    class IfBeforeRule(Rule):
        def __init__(self, if_expr, before, is_ended):
            self.if_expr = if_expr
            self.before = before
            self.is_ended = is_ended

        def before_handle(self, request, kwargs):
            return self.before.exec(request, kwargs)

    class IfAfterRule(Rule):
        def __init__(self, if_expr, after, is_ended):
            self.if_expr = if_expr
            self.after = after
            self.is_ended = is_ended

        def after_handle(self, response, geoip, ipaddr, run):
            return self.after.exec(response, geoip, ipaddr, run)

    class IfBeforeAfterRule(IfBeforeRule, IfAfterRule):
        def __init__(self, if_expr, before, after, is_ended):
            self.if_expr = if_expr
            self.before = before
            self.after = after
            self.is_ended = is_ended

    def __init__(self, next, *rules):
        super().__init__(next)
        self.if_parser = RuleIfParser()
        self.then_parser = RuleThenParser()
        self.before_parser = RuleBeforeParser()
        self.after_parser = RuleAfterParser()
        self.rules = [self._parse_rule(rule) for rule in rules]
        self.geoip_reader = self.server.open_geoip()

    def _parse_rule(self, rule):
        if 'end' not in rule:
            rule['end'] = False
        try:
            len_of_rule = len(rule)
            if len_of_rule == 3:
                if 'then' in rule:
                    return self.IfThenRule(
                        self.if_parser.parse(rule['if']),
                        self.then_parser.parse(rule['then']),
                        rule['end']
                    )
                elif 'before' in rule:
                    return self.IfBeforeRule(
                        self.if_parser.parse(rule['if']),
                        self.before_parser.parse(rule['before']),
                        rule['end']
                    )
                else:
                    return self.IfAfterRule(
                        self.if_parser.parse(rule['if']),
                        self.after_parser.parse(rule['after']),
                        rule['end']
                    )
            elif len_of_rule == 4:
                return self.IfBeforeAfterRule(
                    self.if_parser.parse(rule['if']),
                    self.before_parser.parse(rule['before']),
                    self.after_parser.parse(rule['after']),
                    rule['end']
                )
            raise KeyError()
        except pyparsing.exceptions.ParseException as exc:
            raise InvalidConfig('Invalid rules config: %s' % (exc, ))
        except KeyError:
            raise InvalidConfig('A rule must match [if/then] or [if/before/after] pattern')

    def _geoip(self, ip):
        result = self.geoip_reader.get(ip)                            
        return result['country']['iso_code'] if result else None

    async def _query(self, request, **kwargs):
        response = await super().handle(request, **kwargs)
        ips = map(lambda r: str(r.rdata), response.rr)
        geoips = map(self._geoip, ips)
        return response, ips, geoips

    async def handle(self, request, **kwargs):
        if request.q.qtype not in (dnslib.QTYPE.AAAA, dnslib.QTYPE.A):
            return await super().handle(request, **kwargs)
        
        cached_geoips = dict()
        cached_ipaddrs = dict()
        def geoip(record):
            ip_str = str(record.rdata)
            if ip_str in cached_geoips:
                return cached_geoips[ip_str]
            cached_geoips[ip_str] = country_code = self._geoip(ip_str)
            return country_code
        
        def ipaddr(record):
            ip_str = str(record.rdata)
            if ip_str in cached_ipaddrs:
                return cached_ipaddrs[ip_str]
            if record.rtype == dnslib.QTYPE.A:
                return ipaddress.IPv4Address(ip_str)
            elif record.rtype == dnslib.QTYPE.AAAA:
                return ipaddress.IPv6Address(ip_str)

        def run(command):
            logger.debug('Shell: %s', command)
            return self.server.create_task(
                asyncio.create_subprocess_shell(
                    command, 
                    stdout=subprocess.DEVNULL, 
                    stderr=subprocess.DEVNULL
                ), 
                command
            )

        response = None
        after_handlers = AfterHandlerList()
        domain = request.q.qname.idna().rstrip('.')
        for rule in self.rules:
            if not rule.test(domain): continue

            rule_type = type(rule)
            if rule_type == self.IfThenRule:
                if response is None:
                    response = rule.exec(request)
            elif rule_type == self.IfBeforeRule:
                if response is None:
                    after_handlers.join(rule.before_handle(request, kwargs))
            elif rule_type == self.IfAfterRule:
                after_handlers.join(rule)
            elif rule_type == self.IfBeforeAfterRule:
                if response is None:
                    after_handlers.join(rule.before_handle(request, kwargs))
                after_handlers.join(rule)
            if rule.is_ended: break

        if response is None:
            response = await super().handle(request, **kwargs)
        return after_handlers.after_handle(response, geoip, ipaddr, run)



import asyncio
import base64
import io
import json
import logging
import email
import email.policy
import mimetypes
import html
import pathlib
import random
import ipaddress
import dnslib

from collections import namedtuple
from http import HTTPStatus, HTTPMethod
from http.client import HTTPMessage
from urllib.parse import urlencode, urlsplit, parse_qs, quote

from .scheme import Scheme
from .exceptions import HttpException, HttpHeaderTooLarge, HttpPayloadTooLarge, HttpNotFound, InvalidConfig


HTTP_VERSION = 'HTTP/1.1'

DEFAULT_HTTP_PORT = 80

DEFAULT_HTTPS_PORT = 443


logger = logging.getLogger(__name__)

ContentType = namedtuple('ContentType', ['media_type', 'charset', 'boundary', 'name'])

ContentDisposition = namedtuple('ContentDisposition', ['type', 'name', 'filename'])

UploadedFile = namedtuple('UploadedFile', ['filename', 'name', 'content_type', 'content'])


class HttpMessage(HTTPMessage):
    def get_content_maintype(self):
        maintype = super().get_content_maintype()
        if maintype == 'multipart':
            return ''
        return maintype

    def set_header(self, name, value, **params):
        if self.get(name):
            del self[name]
        return self.add_header(name, str(value), **params)


class FormData:
    def __init__(self):
        self.data = []
        self.boundary = self._generate_boundary()

    def _generate_boundary(self):
        alphabet = '0123456789qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM'
        extra_tail = ''.join(random.sample(alphabet, 16))
        return '----WebKitFormBoundary' + extra_tail

    def get_boundary(self):
        return self.boundary

    def append(self, name, value):
        self.data.append((name, value))
        return self

    def as_bytes(self):
        with io.BytesIO() as fp:
            for item in self.data:
                key, value = item
                fp.write(b'--')
                fp.write(self.boundary.encode())
                fp.write(b'\r\n')
                fp.write(b'Content-Disposition: form-data; name="%s"\r\n' %(html.escape(key).encode(), ))
                fp.write(b'\r\n')
                fp.write(html.escape(value).encode())
                fp.write(b'\r\n')
            fp.write(b'--')
            fp.write(self.boundary.encode())
            fp.write(b'--\r\n')
            fp.seek(0)
            return fp.read(-1)


class Request:
    def get(self, name, default=None):
        if self.query and name in self.query:
            return self.query[name].pop()
        if self.form and name in self.form:
            return self.form[name].pop()
        return default

    def get_all(self, name):
        if self.query and name in self.query:
            return self.query[name]
        if self.form and name in self.form:
            return self.form[name]
        return []

    @property
    def body(self):
        return self._body

    def _parse_post_data(self, body):
        raw_content_type = self.headers.get('Conetent-Type')
        with io.BytesIO() as fp:
            fp.write(b'Content-Type: ')
            fp.write(raw_content_type.encode())
            fp.write(b'\r\n\r\n')
            fp.write(body)
            form_data = email.message_from_binary_file(
                fp, _class=HttpMessage, policy=email.policy.HTTP).get_payload()
        if not form_data: return
        form = dict()
        files = dict()
        for item in form_data:
            content_disposition = parse_content_disposition(item.get('Content-Disposition'))
            if content_disposition.type == 'form-data' and content_disposition.name:
                if content_disposition.filename:
                    uploaded_file = parse_uploaded_file(item, content_disposition)
                    if content_disposition.name in files:
                        files[content_disposition.name].append(uploaded_file)
                    else:
                        files[content_disposition.name] = [uploaded_file]
                else:
                    if content_disposition.name in form:
                        form[content_disposition.name].append(item.get_payload())
                    else:
                        form[content_disposition.name] = [item.get_payload()]
        self.form = form
        self.files = files

    @body.setter
    def body(self, value):
        if value is None: return
        if not isinstance(value, bytes):
            raise TypeError('Response body must be bytes type, %s given' % (type(value), ))

        raw_content_type = self.headers.get('Conetent-Type')
        content_type = parse_content_type(raw_content_type)
        media_type = content_type.media_type
        charset = content_type.charset
        if media_type == 'applicaton/x-www-urlencoded':
            decoded_body = value.decode(charset) if charset \
                else value.decode(charset)
            self.form = parse_qs(decoded_body)
        elif media_type == 'multipart/form-data':
            self._parse_post_data(value)
        self._body = value

    def __init__(self):
        self.method = None
        self._url = None
        self.version = HTTP_VERSION
        self.headers = HttpMessage(email.policy.HTTP)
        self._body = None
        self.query_string = None
        self.path = None
        self.query = None
        self.form = None
        self.files = None

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        parsed_url = urlsplit(value)
        self._url = value
        self.path = quote(parsed_url.path)
        if parsed_url.query:
            self.query_string = parsed_url.query
            self.query = parse_qs(parsed_url.query)
        else:
            self.query_string = ''
            self.query = dict

    def __repr__(self):
        if self.method is None or self.url is None:
            return '<%s>' % self.__class__.__name__
        return '<%s: %s %r>' % (
            self.__class__.__name__,
            self.method,
            self.url,
        )


class Response:
    def __init__(self, status=200):
        self._status = HTTPStatus(status)
        self.reason = self._status.name
        self.version = HTTP_VERSION
        self.headers = HttpMessage(email.policy.HTTP)
        self.body = None

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        if not isinstance(value, HTTPStatus):
            value = HTTPStatus(value)
        self._status = value
        self.reason = value.name

    def __repr__(self):
        return "<%(cls)s status=%(status)d type=%(type)s length=%(len)s>" % {
            'cls': self.__class__.__name__,
            'status': self.status,
            'len': len(self.body),
            'type': self.headers.get('Content-Type'),
        }


class FileResponse(Response):
    def __init__(self, file):
        super().__init__()
        self.file = file

    @property
    def size(self):
        return self.file.stat().st_size

    def __repr__(self):
        return "<%(cls)s status=%(status)d type=%(type)s length=%(len)s>" % {
            'cls': self.__class__.__name__,
            'status': self.status,
            'len': self.size,
            'type': self.headers.get('Content-Type'),
        }


class JsonResponse(Response):
    def __init__(self, payload):
        super().__init__(200)
        self.body = json.dumps(payload).encode()
        self.headers.add_header('Content-Type', 'application/json', charset='utf-8')

    def __repr__(self):
        return "<%(cls)s status=%(status)d type=%(type)s length=%(len)s>" % {
            'cls': self.__class__.__name__,
            'status': self.status,
            'len': len(self.body),
            'type': self.headers.get('Content-Type'),
        }


class Client:
    def __init__(self, hostname, conn):
        self.hostname = hostname
        self.conn = conn

    def prepare_headers(self, headers):
        return headers + [
            ('Host', self.hostname),
            ('Accept-Encoding', 'identity'),
            ('Connection', 'keep-alive'),
        ]

    def build_request(self, method, path, query, headers, body=None):
        req = Request()
        req.method = method
        req.url = '%s?%s' % (path, urlencode(query)) if query else path
        for hdr in headers:
            req.headers.add_header(*hdr)
        if isinstance(body, dict):
            req.headers.set_header('Content-Type', 'application/x-www-form-urlencoded')
            body = urlencode(body).encode()
        elif isinstance(body, FormData):
            req.headers.set_header('Content-Type', 'multipart/form-data', boundary=body.get_boundary())
            body = body.as_bytes()
        req.body = body
        return req

    async def _request(self, request):
        writer = self.conn.writer
        start_line = '%s %s %s\r\n' % (request.method, request.url, request.version)
        writer.write(start_line.encode())
        if request.body:
            request.headers.set_header('Content-Length', len(request.body))
            writer.write(request.headers.as_bytes())
            writer.write(request.body)
        else:
            del request.headers['Content-Length']
            writer.write(request.headers.as_bytes())

        try:
            await writer.drain()
        except Exception as exc:
            raise HttpException(str(exc))

    async def _read_response(self):
        resp = Response()
        try:
            start_line, resp.headers, resp.body = await _read_http_message(self.conn.reader)
            resp.version, status, resp.reason = start_line.split(' ', 2)
            resp.status = HTTPStatus(int(status))
        except (HttpException, asyncio.CancelledError) as exc:
            raise exc
        except:
            raise HttpException('Invalid response')
        return resp

    async def get(self, path, query=None, headers=[]):
        request = self.build_request(HTTPMethod.GET, path, query, self.prepare_headers(headers))
        logger.debug('Request sent to "%s": %s', self.hostname, request)
        await self._request(request)
        response = await self._read_response()
        logger.debug('Response received from "%s": %s', self.hostname, response)
        return response

    async def post(self, path, query=None, headers=[], body=None):
        request = self.build_request(HTTPMethod.POST, path, query, self.prepare_headers(headers), body)
        logger.debug('Request sent to "%s": %s', self.hostname, request)
        await self._request(request)
        response = await self._read_response()
        logger.debug('Response received from "%s": %s', self.hostname, response)
        return response


class Server:
    MAX_REQUEST_SIZE = 1024 * 1024 * 16

    DEFAULT_FILES = ['index.html', 'index.htm', 'default.htm', 'default.html']

    def __init__(self, config, dispatcher, loop=None):
        self.loop = asyncio.get_running_loop() if loop is None else loop
        self.config = config
        self.dispatcher = dispatcher
        self._server = None
        logger.debug('HTTP serivce initialized')

    async def _respond(self, writer, response):
        try:
            response.headers.add_header('Server', 'DNSPooh')
            start_line = '%s %s %s\r\n' % (response.version, response.status, response.status.name)
            writer.write(start_line.encode())
            if isinstance(response, FileResponse):
                ctype, charset = mimetypes.guess_type(response.file)
                kwargs = {} if charset is None else {'charset': charset}
                response.headers.set_header(
                    'Content-Type', 
                    'application/octet-stream' if ctype is None else ctype,
                    **kwargs
                )
                with response.file.open('rb') as fp:
                    response.headers.set_header('Content-Length', response.size)
                    writer.write(response.headers.as_bytes())
                    await writer.drain()
                    await self.loop.sendfile(writer.transport, fp)
                return response
            elif response.body:
                response.headers.set_header('Content-Length', len(response.body))
                writer.write(response.headers.as_bytes())
                writer.write(response.body)
            else:
                del response.headers['Content-Length']
                writer.write(response.headers.as_bytes())
            await writer.drain()
        except:
            raise IOError('Response begin sending and an error occurred')
        return response
        
    async def _read_request(self, reader):
        req = Request()
        try:
            start_line, req.headers, req.body = await _read_http_message(reader, self.MAX_REQUEST_SIZE)
            method, req.url, req.version = start_line.split(' ', 2)
            req.method = HTTPMethod(method)
        except HttpException as exc:
            raise exc
        except:
            raise HttpException('Invalid request')
        return req

    def _resolve_static_file(self, path):
        file_path = self.static_files.joinpath('.' + path)

        if file_path.is_dir():
            for default_file in self.DEFAULT_FILES:
                file_path = file_path.joinpath(default_file)
                if file_path.exists():
                    return file_path
        elif file_path.exists():
            return file_path

        return False

    async def on_request(self, request):
        static_file = self._resolve_static_file(request.path)
        if static_file:
            return FileResponse(static_file)
        return self.dispatcher.dispatch(request)

    async def on_error(self, status):
        resp = Response(status)
        resp.headers.add_header('Connection', 'close')
        return resp

    async def on_connect(self, reader, writer):
        peername = writer.transport.get_extra_info('peername')
        logger.debug('Connection from %s:%d' % peername)
        while not writer.transport.is_closing():
            try:
                request = await asyncio.wait_for(self._read_request(reader), self.timeout)
                logger.debug('Request received from "%s:%d": %s', *peername, request)
                response = await self._respond(writer, await self.on_request(request))
                logger.debug('Response sent to "%s:%d": %s', *peername, response)
            except HttpException:
                await self._respond(writer, await self.on_error(400))
                writer.transport.close()
                break
            except (TimeoutError, EOFError, IOError):
                writer.transport.close()
                break
            except Exception as exc:
                await self._respond(writer, await self.on_error(500))
                writer.transport.close()
                logger.warning('Server error: %s', exc)
                break

    async def run(self):
        http_config = self.config['http']
        if http_config.get('disable'):
            return
        static_files = pathlib.Path(http_config['static_files'])
        if not static_files.is_dir():
            raise InvalidConfig('%s is not a directory' % (static_files, ))
        self.static_files = static_files
        host = http_config['host']
        port = int(http_config['port'])
        if host != '127.0.0.1' and host != 'localhost':
            logger.warn('HTTP server host %s is not safe', host)
        self.timeout = http_config['timeout']
        self._server = await asyncio.start_server(self.on_connect, host, port)
        logger.info('HTTP serivce started')
        logger.info('HTTP server is available at http://%s:%d/', host, port)

        try:
            async with self._server:
                await self._server.serve_forever()
        except asyncio.CancelledError:
            logger.debug('HTTP serivce interrupted')
        finally:
            self._server.close()
            #self.status = self.Status.stopped
            logger.info('HTTP serivce stopped')


async def _read_http_message(reader, max_body=None):
    try:
        http_header = await reader.readuntil(b'\r\n\r\n')
    except asyncio.exceptions.LimitOverrunError as exc:
        raise HttpHeaderTooLarge(exc)
    with io.StringIO(http_header.decode()) as fp:
        start_line = fp.readline()
        headers = email.message_from_file(fp, _class=HttpMessage, policy=email.policy.HTTP)
    body = None
    if headers.get('Transfer-Encoding') == 'chunked':
        with io.BytesIO() as fp:
            while True:
                chunk_len_ln = await reader.readuntil(b'\r\n')
                chunk_len = int(chunk_len_ln[:-2].decode(), 16)
                if chunk_len == 0: break
                fp.write(await reader.readexactly(chunk_len))
                if b'\r\n' != await reader.readexactly(2):
                    raise HttpException('Chunked encoding error, missing CRLF')
                if max_body and fp.tell() > max_body:
                    raise HttpPayloadTooLarge('The size of payload is greater than %d', max_body)
            fp.seek(0)
            body = fp.read(-1)
    else:
        content_length = headers['Content-Length']
        if content_length is not None:
            with io.BytesIO() as fp:
                body = await reader.readexactly(int(content_length))
        elif headers.get('Connection') == 'close':
            body = await reader.read()
    return start_line, headers, body


def parse_content_type(content_type):
    if not content_type:
        return ContentType(None, None, None, None)
    directives = content_type.split(';')
    media_type = directives.pop(0).strip().lower()
    subitems = dict([[__.strip().lower() for __ in _.split('=', 1)] for _ in directives])

    return ContentType(media_type, 
                       subitems.get('charset'), 
                       subitems.get('boundary'), 
                       subitems.get('name'))


def parse_content_disposition(content_disposition):
    if not content_disposition:
        return ContentDisposition(None, None, None)
    directives = content_disposition.split(';')
    _type = directives.pop(0).strip().lower()
    subitems = dict([[__.strip().lower() for __ in _.split('=', 1)] for _ in directives])

    return ContentType(_type, 
                       subitems.get('name'),  
                       subitems.get('filename'))


def parse_uploaded_file(message, content_disposition=None):
    if content_disposition is None:
        content_disposition = parse_content_disposition(message.get('Content-Disposition'))
    name = html.unescape(content_disposition.name)
    filename = html.unescape(content_disposition.filename)
    encoding = message.get('Content-Transfer-Encoding')
    content_type = parse_content_type(message.get('Conent-Type'))
    content = message.get_payload()
    if encoding == 'base64':
        content = base64.b64decode(content)
    else:
        content = content.encode()
    return UploadedFile(filename, name, content_type, content)


class Dispatcher:
    def __init__(self, dns_server):
        super().__init__()
        self.dns_server = dns_server

    def _status(self):
        return JsonResponse({
            'status': self.dns_server.status.name,
            'upstreams': [vars(up) for up in self.dns_server.upstreams],
            'stats': [record.as_dict() for record in self.dns_server.stats.records],
        })

    def _stop(self, request):
        pass

    def _restart(self, request):
        pass

    def dispatch(self, request):
        match request.method, request.path:
            case HTTPMethod.GET, '/status': return self._status()
            case HTTPMethod.POST, '/stop': return self._stop(request)
            case HTTPMethod.POST, '/restart': return self._start(request)
        raise HttpNotFound()


async def fetch(url, resolver, pool, proxy=None, **kwargs):
    parsed_url = urlsplit(url)
    hostname = parsed_url.hostname
    if parsed_url.scheme == 'http':
        scheme = Scheme.tcp
        port = parsed_url.port if parsed_url.port else DEFAULT_HTTP_PORT
    elif parsed_url.scheme == 'https':
        scheme = Scheme.tls
        port = parsed_url.port if parsed_url.port else DEFAULT_HTTPS_PORT
    else:
        raise ValueError('Invalid url scheme %s' % (parsed_url.scheme, ))
    try:
        ipaddress.ip_address(hostname)
        host = hostname
    except ValueError:
        dns_request = dnslib.DNSRecord.question(hostname)
        dns_response = await resolver(dns_request)
        if not dns_response or dns_response.header.a == 0:
            raise HttpException('Could not resolve domain %s' % (hostname, ))
        host = str(dns_response.rr[0].rdata)
    if 'method' in kwargs:
        method =HTTPMethod(kwargs['method'])
    else:
        method = HTTPMethod.GET
    if method not in (HTTPMethod.GET, HTTPMethod.POST):
        raise ValueError('Unsupported http method %s' % (method.name, ))
    url_path = quote(parsed_url.path)
    if parsed_url.query:
        url_path = '%s?%s' % (url_path, parsed_url.query)
    headers = kwargs.get('headers', [])

    with await pool.connect(host, port, scheme, proxy, pooled=False) as conn:
        if method == HTTPMethod.GET:
            return await Client(hostname, conn).get(url_path, headers=headers)
        else:
            body = kwargs.get('body')
            return await Client(hostname, conn).post(url_path, headers=headers, body=body)



import unittest
import ipaddress

import dnslib

from dnspooh.middlewares.rules import RuleIfParser, RuleThenParser, RuleBeforeParser, RuleAfterParser, ReplacDomainHandler


if_parser = RuleIfParser()
then_parser = RuleThenParser()
before_parser = RuleBeforeParser()
after_parser = RuleAfterParser()


def dns_request(domain):
    return dnslib.DNSRecord.question(domain)


def dns_answer(request, ip_addr):
    request.add_answer(dnslib.RR(request.q.qname, dnslib.QTYPE.A,
        rdata=dnslib.A(ip_addr)))
    return request


def dns_response(domain, *ip_addrs):
    request = dns_request(domain)
    for ip_addr in ip_addrs:
        dns_answer(request, ip_addr)
    return request


class RuleIfTest(unittest.TestCase):
    def test_in_domain(self):
        self.assertTrue(if_parser.parse('qq in domain').test('weixin.qq.com'))
        self.assertTrue(if_parser.parse('(qq, weibo) in domain').test('weixin.qq.com'))
        self.assertFalse(if_parser.parse('(qq, weibo) in domain').test('taobao.com'))

    def test_not_in_domain(self):
        self.assertFalse(if_parser.parse('qq not in domain').test('weixin.qq.com'))
        self.assertFalse(if_parser.parse('(qq, weibo) not in domain').test('weixin.qq.com'))
        self.assertTrue(if_parser.parse('(qq, weibo) not in domain').test('taobao.com'))

    def test_domain_is(self):
        self.assertTrue(if_parser.parse('domain is qq.com').test('qq.com'))
        self.assertTrue(if_parser.parse('domain is (baidu.com, taobao.com, weibo.com, qq.com)').test('qq.com'))
        self.assertFalse(if_parser.parse('domain is (baidu.com, taobao.com, weibo.com, qq.com)').test('tencent.com'))

    def test_domain_is_not(self):
        self.assertFalse(if_parser.parse('domain is not qq.com').test('qq.com'))
        self.assertFalse(if_parser.parse('domain is not (baidu.com, taobao.com, weibo.com, qq.com)').test('qq.com'))
        self.assertTrue(if_parser.parse('domain is not (baidu.com, taobao.com, weibo.com, qq.com)').test('tencent.com'))

    def test_domain_starts_with(self):
        self.assertTrue(if_parser.parse('domain starts with www').test('www.qq.com'))
        self.assertTrue(if_parser.parse('domain starts with (mobile, www, test)').test('www.qq.com'))
        self.assertFalse(if_parser.parse('domain starts with (mobile, www, test)').test('tencent.com'))

    def test_domain_starts_without(self):
        self.assertFalse(if_parser.parse('domain starts without www').test('www.qq.com'))
        self.assertFalse(if_parser.parse('domain starts without (mobile, www, test)').test('www.qq.com'))
        self.assertTrue(if_parser.parse('domain starts without (mobile, www, test)').test('tencent.com'))

    def test_domain_ends_with(self):
        self.assertTrue(if_parser.parse('domain ends with .com').test('www.qq.com'))
        self.assertTrue(if_parser.parse('domain ends with (.com, .cn, .com.cn)').test('www.qq.com'))
        self.assertFalse(if_parser.parse('domain ends with (.com, .cn, .com.cn)').test('mozilla.org'))

    def test_domain_ends_without(self):
        self.assertFalse(if_parser.parse('domain ends without .com').test('www.qq.com'))
        self.assertFalse(if_parser.parse('domain ends without (.com, .cn, .com.cn)').test('www.qq.com'))
        self.assertTrue(if_parser.parse('domain ends without (.com, .cn, .com.cn)').test('mozilla.org'))

    def test_domain_match(self):
        self.assertTrue(if_parser.parse('domain match /w+\.baidu\.com/').test('www.baidu.com'))
        self.assertFalse(if_parser.parse('domain match /w+\.baidu\.com/').test('baidu.com'))

    def test_not_and_or(self):
        self.assertTrue(if_parser.parse('domain is qq.com or domain ends with baidu.com and not qq in domain').test('www.baidu.com'))


class RuleThenTest(unittest.TestCase):
    def test_block(self):
        response = then_parser.parse('block').exec(dns_request('qq.com'))
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_return(self):
        response = then_parser.parse('return 127.0.0.1').exec(dns_request('qq.com'))
        record = response.rr[0]
        self.assertEqual(record.rtype, dnslib.QTYPE.A)
        self.assertEqual(str(record.rdata), '127.0.0.1')
        response = then_parser.parse('return (127.0.0.1, 127.0.0.2, 127.0.0.3)').exec(dns_request('qq.com'))
        self.assertEqual(response.header.a, 3)
        self.assertEqual(str(response.rr[0].rdata), '127.0.0.1')
        self.assertEqual(str(response.rr[1].rdata), '127.0.0.2')
        self.assertEqual(str(response.rr[2].rdata), '127.0.0.3')


class RuleBeforeTest(unittest.TestCase):
    def test_replace_domain(self):
        kwargs = dict()
        request = dns_request('baidu.com')
        after_handler = before_parser.parse('replace domain by google.com').exec(request, kwargs)
        self.assertTrue(isinstance(after_handler, ReplacDomainHandler))
        self.assertEqual(request.q.qname, 'google.com')
        response = after_handler.after_handle(dns_answer(request, '127.0.0.1'), None, None, None)
        self.assertEqual(response.q.qname, 'baidu.com')

    def test_set_upstream(self):
        kwargs = dict()
        request = dns_request('baidu.com')
        after_handler = before_parser.parse('set upstream group to cn, set upstream name to cloudflare-tls').exec(request, kwargs)
        self.assertEqual(after_handler, None)
        self.assertEqual(kwargs['upstream_group'], 'cn')
        self.assertEqual(kwargs['upstream_name'], 'cloudflare-tls')


def geoip(code):
    def _geoip(_):
        return code
    return _geoip


def ipaddr(record):
    ip_str = str(record.rdata)
    if record.rtype == dnslib.QTYPE.A:
        return ipaddress.IPv4Address(ip_str)
    elif record.rtype == dnslib.QTYPE.AAAA:
        return ipaddress.IPv6Address(ip_str)

def run(command):
    print('shell:', command)
    return

class RuleAfterTest(unittest.TestCase):
    def test_block_if(self):
        response = after_parser.parse('block if always').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_return_if(self):
        response = after_parser.parse('return 127.0.0.10 if always').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        record = response.rr[0]
        self.assertEqual(record.rtype, dnslib.QTYPE.A)
        self.assertEqual(str(record.rdata), '127.0.0.10')
        response = after_parser.parse('return (127.0.0.1, 127.0.0.2, 127.0.0.3) if always').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.a, 3)
        self.assertEqual(str(response.rr[0].rdata), '127.0.0.1')
        self.assertEqual(str(response.rr[1].rdata), '127.0.0.2')
        self.assertEqual(str(response.rr[2].rdata), '127.0.0.3')
    
    def test_add_record(self):
        response = after_parser.parse('add record (127.0.0.3, 127.0.0.4) if always').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        response = after_parser.parse('add record 127.0.0.5 if always').exec(response, geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.a, 5)
        self.assertEqual(str(response.rr[0].rdata), '127.0.0.1')
        self.assertEqual(str(response.rr[1].rdata), '127.0.0.2')
        self.assertEqual(str(response.rr[2].rdata), '127.0.0.3')
        self.assertEqual(str(response.rr[3].rdata), '127.0.0.4')
        self.assertEqual(str(response.rr[4].rdata), '127.0.0.5')

    def test_remove_record(self):
        response = after_parser.parse('remove record where ip is 127.0.0.1').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.a, 1)
        self.assertEqual(str(response.rr[0].rdata), '127.0.0.2')

    def test_replace_record(self):
        response = after_parser.parse('replace record by 192.168.1.1 where ip is 127.0.0.1').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.a, 2)
        self.assertEqual(str(response.rr[0].rdata), '192.168.1.1')
        self.assertEqual(str(response.rr[1].rdata), '127.0.0.2')

    def test_any_ip_is(self):
        response = after_parser.parse('block if any ip is 127.0.0.1').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        response = after_parser.parse('block if any ip is (127.0.0.1, 192.168.1.2)').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_any_ip_is_not(self):
        response = after_parser.parse('block if any ip is not 127.0.0.1').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        response = after_parser.parse('block if any ip is not (127.0.0.1, 127.0.0.3)').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_all_ip_is(self):
        response = after_parser.parse('block if all ip is 127.0.0.1').exec(dns_response('qq.com', '127.0.0.1'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        response = after_parser.parse('block if all ip is (127.0.0.1, 127.0.0.2)').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_all_ip_is_not(self):
        response = after_parser.parse('block if all ip is not 192.168.1.1').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        response = after_parser.parse('block if all ip is not (192.168.1.1, 192.168.1.2)').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_any_ip_in(self):
        response = after_parser.parse('block if any ip in 127.0.0.1/24').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        response = after_parser.parse('block if any ip in (127.0.0.1/24, 192.168.1.2/24)').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_any_ip_not_in(self):
        response = after_parser.parse('block if any ip not in 127.0.0.1/24').exec(dns_response('qq.com', '127.0.0.1', '172.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        response = after_parser.parse('block if any ip not in (127.0.0.1/24, 192.168.1.1/24)').exec(dns_response('qq.com', '127.0.0.1', '172.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_all_ip_in(self):
        response = after_parser.parse('block if all ip in 127.0.0.1/24').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        response = after_parser.parse('block if all ip in (127.0.0.1/24, 172.0.0.2/24)').exec(dns_response('qq.com', '127.0.0.1', '172.0.0.1'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_all_ip_not_in(self):
        response = after_parser.parse('block if all ip not in 192.168.1.1/24').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        response = after_parser.parse('block if all ip not in (192.168.1.1/24, 192.168.253.1/24)').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_geoip_is(self):
        response = after_parser.parse('block if any geoip is cn').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        response = after_parser.parse('block if all geoip is cn').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_geoip_is_not(self):
        response = after_parser.parse('block if any geoip is not cn').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('en'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)
        response = after_parser.parse('block if all geoip is not cn').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2'), geoip('en'), ipaddr, run)
        self.assertEqual(response.header.rcode, dnslib.RCODE.NXDOMAIN)

    def test_first_last(self):
        response = after_parser.parse('remove record where first or last').exec(dns_response('qq.com', '127.0.0.1', '127.0.0.2', '127.0.0.3'), geoip('cn'), ipaddr, run)
        self.assertEqual(response.header.a, 1)
        self.assertEqual(str(response.rr[0].rdata), '127.0.0.2')



import argparse
import asyncio
import logging
import sys

from . import https
from . import server
from . import __version__
from .config import *
from .upstream import *


logger = logging.getLogger(__name__)


def parse_arguments():
    parser = argparse.ArgumentParser(
        prog = 'DNSPooh',
        description = 'A MitM DNS Proxy',
        add_help=False
    )

    parser.add_argument('-c', '--config', metavar='file', dest='config',
                        help='config file path (example "%s")' % (CONFIG_FILE, ))
    parser.add_argument('-u', '--upstream', metavar='servers', dest='upstreams', nargs='+',
                        help='space-separated upstream DNS servers list')
    parser.add_argument('-t', '--timeout', metavar='ms', dest='timeout', type=int, 
                        help='milliseconds for upstream DNS response timeout (default %d ms)' % (UPSTREAM_TIMEOUT, ))
    parser.add_argument('-h', '--host', metavar='host', dest='host', 
                        help='local DNS proxy server listening host (default "%s")' % (LISTEN_HOST, ))
    parser.add_argument('-p', '--port', metavar='port', dest='port', type=int, 
                        help='local DNS proxy server listening port (default "%s")' % (DEFAULT_DNS_PORT, ))
    parser.add_argument('-D', '--debug', action='store_true', help='display debug message')
    parser.add_argument('-d', '--dump', action='store_true', default=False, help='dump pretty config data')
    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + __version__)
    parser.add_argument('--help', action='help', help='show this help message and exit')

    return parser.parse_args(sys.argv[1:])


async def startup():
    try:
        args = parse_arguments()
        config = Config.load(args)
        if args.dump:
            from pprint import pprint
            pprint(config.conf)
            return

        debug = config['debug']
        if debug:
            logging.root.setLevel(logging.DEBUG)

        loop = asyncio.get_running_loop()
        loop.set_debug(debug)

        dns_server = server.Server(config, loop)
        dispatcher = https.Dispatcher(dns_server)
        http_server = https.Server(config, dispatcher, loop)

        await asyncio.gather(dns_server.run(), http_server.run())
    except asyncio.CancelledError:
        logger.info('Exit')
    except Exception as exc:
        logger.error(exc)


def main():
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(name)s.%(levelname)s] %(message)s",
        handlers=[logging.StreamHandler()]
    )
    asyncio.run(startup())



import enum
import time

from collections import deque

from .exceptions import EmptyValueError, UnexpectedValueError


class Stats:
    class Record:
        class Result(enum.Enum):
            success = enum.auto()
            timeout = enum.auto()
            error = enum.auto()
            empty = enum.auto()
            invalid = enum.auto()

            def __repr__(self):
                return self.name

        def __init__(self, stats, upstream):
            self.stats = stats
            self.upstream = upstream

        def __enter__(self):
            self.start_time = time.time()
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            self.upstream.usage += 1
            if exc_type is None:
                self.upstream.success += 1
                self.result = self.Result.success
            else:
                if isinstance(exc_type, TimeoutError):
                    self.result = self.Result.timeout
                elif isinstance(exc_type, EmptyValueError):
                    self.result = self.Result.empty
                elif isinstance(exc_type, UnexpectedValueError):
                    self.result = self.Result.invalid
                else:
                    self.result = self.Result.error

            self.cost_time = time.time() - self.start_time
            self.stats.records.append(self)
            del self.stats

        def __repr__(self):
            return str(vars(self))

        def as_dict(self):
            return {
                'upstream': self.upstream.name,
                'start_time': self.start_time,
                'cost_time': self.cost_time,
            }

    def __init__(self, maxlen):
        self.records = deque(maxlen=maxlen)

    def record(self, upstream):
        return self.Record(self, upstream)

    def __repr__(self):
        return str(vars(self))




def split_domain(name, reverse=True):
    dot_pos = name.find('.')
    if reverse:
        yield name
        while dot_pos > 0:
            yield name[dot_pos + 1:]
            dot_pos = name.find('.', dot_pos + 1)
    else:
        while dot_pos > 0:
            yield name[:dot_pos]
            dot_pos = name.find('.', dot_pos + 1)
        yield name



import base64
import struct
import ipaddress

from urllib.parse import urlsplit

from .scheme import Scheme


DEFAULT_HTTP_PROXY_PORT = 8080

DEFAULT_SOCKS5_PROXY_PORT = 1080

class Proxy:
    def __init__(self, url, hostname, port, host=None, username=None, password=None):
        self.url = url
        self.hostname = hostname
        self.port = port
        self.host = host
        self.username = username
        self.password = password

    def __repr__(self):
        return self.url

    def has_auth(self):
        return self.username is not None and \
               self.password is not None

    def udp_tunnel_enabled(self):
        return False

    async def handshake(self, reader, writer, remote_addr):
        raise NotImplementedError()


class HttpProxy(Proxy):
    async def handshake(self, reader, writer, remote_addr):
        if self.has_auth():
            credentials = base64.b64encode(
                ('%s:%s' % (self.username, self.password)).encode()
            ).decode()
            request = 'CONNECT %s:%s HTTP/1.1\r\n' % remote_addr
            request += 'Proxy-Authorization: basic %s\r\n\r\n' % (credentials, )
        else:
            request = 'CONNECT %s:%s HTTP/1.1\r\n\r\n' % remote_addr

        writer.write(request.encode())
        await writer.drain()
        response = await reader.readuntil(b'\r\n\r\n')
        return response.startswith(b'HTTP/1.1 200')


class Socks5Proxy(Proxy):
    VERSION = 5

    AUTH_METHOD = 2

    NONE_METHOD = 0

    AUTH_SUCCESS = 0

    CMD_CONNECT = 1

    CMD_UDP_ASSOCIATE = 3

    ATYP_IPV4 = 1

    ATYP_IPV6 = 4

    REP_SUCCESS = 0

    class UDPTunnel:
        def __init__(self, addr):
            self.addr = addr

        def parse(self, data, src_addr):
            _, _, atype = struct.unpack('!H2B', data[:4])
            if atype == Socks5Proxy.ATYP_IPV4:
                _from_addr, from_port = struct.unpack('!4sH', data[4:10])
                from_addr = (str(ipaddress.IPv4Address(_from_addr)), from_port)
                entity_data = data[10:]
            elif atype == Socks5Proxy.ATYP_IPV6:
                _from_addr, from_port = struct.unpack('!16sH', data[4:22])
                from_addr = (str(ipaddress.IPv6Address(_from_addr)), from_port)
                entity_data = data[22:]
            else:
                raise ValueError('Invalid ATYPE %d received' % atype)
            if from_addr != src_addr:
                raise ValueError('Source address %s:%d does not match' % from_addr)

            return entity_data

        def pack(self, data, dst_addr):
            ip, port = dst_addr
            ip_addr = ipaddress.ip_address(ip)
            if isinstance(ip_addr, ipaddress.IPv4Address):
                pack_header = struct.pack(
                    '!H2B4sH', 0, 0, 
                    Socks5Proxy.ATYP_IPV4,
                    ip_addr.packed, 
                    port
                )
            elif isinstance(ip_addr, ipaddress.IPv6Address):
                pack_header = struct.pack(
                    '!H2B16sH', 0, 0, 
                    Socks5Proxy.ATYP_IPV6,
                    ip_addr.packed, 
                    port
                )
            else:
                raise ValueError('Invalid destination address "%s"' % ip)

            return pack_header + data

    def udp_tunnel_enabled(self):
        return True

    async def _handshake(self, reader, writer, remote_addr, scheme=Scheme.tcp):
        writer.write(struct.pack('!3B', self.VERSION, 1, self.AUTH_METHOD))
        await writer.drain()
        server_version, method = struct.unpack('!2B', await reader.readexactly(2))
        if server_version != self.VERSION:
            raise ConnectionError('Unsupported socks proxy version %d' % (server_version, ))
        if method != self.NONE_METHOD:
            if not self.has_auth():
                raise ConnectionError('Proxy "%s" need authentication' % (self.url, ))
            if method != self.AUTH_METHOD:
                raise ConnectionError('Unsupported socks proxy authentication method %d' % (method, ))

            username = self.username.encode()
            password = self.password.encode()
            username_len = len(username)
            password_len = len(password)
            writer.write(struct.pack(
                '!2B%dsB%ds' % (username_len, password_len), 
                self.VERSION, 
                username_len,
                username,
                password_len,
                password
            ))
            await writer.drain()
            method, status = struct.unpack('!2B', await reader.readexactly(2))
            if status != self.AUTH_SUCCESS:
                raise ConnectionError('Socks proxy authentication failed')

        dst_addr, dst_port = remote_addr
        ip_addr = ipaddress.ip_address(dst_addr)
        if isinstance(ip_addr, ipaddress.IPv4Address):
            writer.write(struct.pack(
                '!4B4sH', 
                self.VERSION, 
                self.CMD_CONNECT if scheme == Scheme.tcp else self.CMD_UDP_ASSOCIATE,
                0, 
                self.ATYP_IPV4,
                ip_addr.packed,
                dst_port
            ))
        elif isinstance(ip_addr, ipaddress.IPv6Address):
            writer.write(struct.pack(
                '!4B16sH', 
                self.VERSION, 
                self.CMD_CONNECT if scheme == Scheme.tcp else self.CMD_UDP_ASSOCIATE,
                0, 
                self.ATYP_IPV6,
                ip_addr.packed,
                dst_port
            ))
        else:
            raise ValueError('Invalid remote address "%s:%d"' % remote_addr)

        await writer.drain()
        _, rep, _, atype,  = struct.unpack('!4B', await reader.readexactly(4))

        if rep != self.REP_SUCCESS:
            raise ConnectionError('Failed to connection remote address "%s:%d"' % remote_addr)

        if atype == self.ATYP_IPV4:
            bind_addr, bind_port = struct.unpack('!4sH', await reader.readexactly(6))
            bind_addr = str(ipaddress.IPv4Address(bind_addr))
        elif atype == self.ATYP_IPV6:
            bind_addr, bind_port = struct.unpack('!16sH', await reader.readexactly(18))
            bind_addr = str(ipaddress.IPv6Address(bind_addr))
        else:
            raise ConnectionError('Invalid response atype')

        if scheme != Scheme.udp and (bind_addr != self.host or bind_port != self.port):
            raise ConnectionError('Relay mode does not supported')

        return bind_addr, bind_port

    async def handshake(self, reader, writer, remote_addr):
        try:
            await self._handshake(reader, writer, remote_addr, Scheme.tcp)
        except ConnectionError:
            return False

        return True

    async def make_udp_tunnel(self, reader, writer, remote_addr):
        return self.UDPTunnel(
            await self._handshake(reader, writer, remote_addr, Scheme.udp)
        )


def parse_proxy(url):
    if not url:
        return None

    parsed_url = urlsplit(url)
    if parsed_url.path != '' and \
        parsed_url.path != '/' or \
        parsed_url.query != '' or \
        parsed_url.fragment != '':
        raise ValueError('Invalid proxy "%s"' % (url, ))

    try:
        ipaddress.ip_address(parsed_url.hostname)
        host = parsed_url.hostname
    except ValueError:
        host = None

    if parsed_url.scheme == 'http':
        return HttpProxy(
            url, 
            parsed_url.hostname, 
            parsed_url.port if parsed_url.port \
                else DEFAULT_HTTP_PROXY_PORT,
            host,
            parsed_url.username,
            parsed_url.password
        )
    elif parsed_url.scheme == 'socks5':
        return Socks5Proxy(
            url, 
            parsed_url.hostname, 
            parsed_url.port if parsed_url.port \
                else DEFAULT_SOCKS5_PROXY_PORT,
            host,
            parsed_url.username,
            parsed_url.password
        )
    else:
        raise ValueError('Invalid proxy scheme "%s" in "%s"' % (parsed_url.scheme, url))



import asyncio
import base64
import logging
import struct
import functools
import enum
import time

from importlib import resources

import maxminddb

from dnslib import DNSRecord, DNSError, DNSHeader

from . import https
from . import middlewares
from .config import DnsUpstream, HttpsUpstream, TlsUpstream
from .pool import Pool
from .scheme import Scheme
from .exceptions import *
from .stats import Stats
from .upstream import UpstreamCollection


logger = logging.getLogger(__name__)


class ServerProtocol(asyncio.DatagramProtocol):
    def __init__(self, server):
        super().__init__()
        self.server = server
        self.need_restart = False

    def datagram_received(self, data, addr):
        self.server.on_request(data, addr)

    def error_received(self, exc):
        if isinstance(exc, OSError):
            logger.debug(exc)
        else:
            logger.info('DNS server error received: %s', exc)
        self.need_restart = True
        self.server.transport.abort()

    def connection_lost(self, exc):
        super().connection_lost(exc)
        if self.need_restart:
            self.need_restart = False
            self.server.on_error_restart()


class QueryProtocol(asyncio.DatagramProtocol):
    def __init__(self, data, response):
        super().__init__()
        self.data = data
        self.response = response

    def connection_made(self, transport):
        self.transport = transport
        transport.sendto(self.data)

    def datagram_received(self, data, addr):
        self.transport.close()
        self.response.set_result(data)

    def error_received(self, exc):
        logger.error('DNS query error: %s', exc)


class Server:
    class Status(enum.Enum):
        initialized = enum.auto()
        start_pedding = enum.auto()
        running = enum.auto()
        restart_pedding = enum.auto()
        stop_pedding = enum.auto()
        stopped = enum.auto()

    def __init__(self, config, loop=None):
        self.config = config
        self.pool = Pool()
        self.loop = asyncio.get_running_loop() if loop is None else loop
        self.abort_event = asyncio.Event()
        self.middlewares = self._create_middlewares()
        self.transport = None
        self.stats = Stats(config['stats.max_size'])
        self.status = self.Status.initialized
        self.tasks = []
        logger.debug('DNS serivce initialized')

    async def bootstrap(self):
        logger.debug('DNS service bootstrapping')
        self.host = self.config['host']
        self.port = self.config['port']
        self.timeout = self.config['timeout']
        self.upstreams = UpstreamCollection(self.config['upstreams'], 
                                            self.config['secure'])
        self.proxy = self.config['proxy']
        if self.proxy:
            logger.info('Using proxy %s', self.proxy)
        bootstrap_upstreams = []
        hostname_upstreams = []
        named_upstreams = dict()

        for upstream in self.upstreams.all():
            if upstream.host:
                bootstrap_upstreams.append(upstream)
            else:
                hostname_upstreams.append(upstream)

            if upstream.name:
                if upstream.name in named_upstreams:
                    raise InvalidConfig('Duplicated upstream name "%s"' % (upstream.name, ))
                named_upstreams[upstream.name] = upstream

        async def resolve_upstream_hostname(hostname_upstream, bootstrap_upstreams):
            request = DNSRecord.question(hostname_upstream.hostname)
            response = await self.handle(request, upstreams=bootstrap_upstreams)
            if not response or response.header.a == 0:
                logger.warning('Failed to resolve upstream domain "%s"', hostname_upstream.hostname)
                hostname_upstream.disable = True
            else:
                hostname_upstream.host = str(response.rr[0].rdata)

        await asyncio.gather(*[
            resolve_upstream_hostname(hostname_upstream, bootstrap_upstreams) \
                for hostname_upstream in hostname_upstreams
        ])

        await self.test_upstreams('google.com')
        return True

    async def test_upstreams(self, hostname):
        request = DNSRecord.question(hostname)
        async def test_upstream(upstream):
            start_counter = time.perf_counter()
            response = await self.handle(request, upstreams=[upstream])
            cost_time = time.perf_counter() - start_counter
            if not response or response.header.a == 0:
                upstream.disable = True
                upstream.priority = -1
                logger.warning('Test upstream %s failed', upstream.name)
            else:
                timeout = self._get_timeout(upstream)
                upstream.priority = int(max(0, timeout - cost_time) / timeout * 1000)
                logger.info('Test upstream %s passed, responding speed: %d', upstream.name, upstream.priority)

        await asyncio.gather(*[test_upstream(_) for _ in self.upstreams.all() if not _.disable])
        self.upstreams.sort()
        primary_upstream = self.upstreams.primary
        if primary_upstream.disable:
            raise NetworkError('No available upstream server')
        logger.info('Primary DNS is %s', primary_upstream.name)

    def create_task(self, coro, name=None, context=None):
        task = self.loop.create_task(coro, name=name, context=context)
        self.tasks.append(task)
        task.add_done_callback(lambda _: self.remove_task(task))
        return task

    def create_scheduled_task(self, coro, timer, name=None, context=None):
        async def _task():
            async for i in timer:
                logger.debug('Schedule task "%s" repeat times %d' % (name, i))
                await coro()

        return self.create_task(_task(), name, context)

    def remove_task(self, task):
        self.tasks.remove(task)

    def _get_proxy(self, upstream):
        return self.proxy if upstream.proxy is None \
            else upstream.proxy

    def _create_middlewares(self):
        wrapped = self
        names = self.config['middlewares']
        for name in names:
            wrapped = middlewares.create_middleware(
                name, wrapped, self.config.get(name)
            )
            logger.info('%s loaded', wrapped.__class__.__name__)
        return wrapped

    async def _resolve_by_dns(self, query, upstream):
        response_future = self.loop.create_future()
        upstream_addr = upstream.to_addr()

        proxy = self._get_proxy(upstream)
        if proxy and proxy.udp_tunnel_enabled():
            try:
                conn = await self.pool.connect(
                    upstream.host, 
                    upstream.port, 
                    Scheme.udp, 
                    proxy
                )
            except ConnectionError as exc:
                logger.warning('Failed to connect to proxy %s: %s', proxy.hostname, exc)
                return
            transport, _ = await self.loop.create_datagram_endpoint(
                lambda: QueryProtocol(
                    conn.udp_tunnel.pack(query, upstream_addr), 
                    response_future
                ),
                remote_addr=conn.udp_tunnel.addr
            )
            try:
                response = conn.udp_tunnel.parse(
                    await response_future, 
                    upstream_addr
                )
            finally:
                transport.close()
            return response

        transport, _ = await self.loop.create_datagram_endpoint(
            lambda: QueryProtocol(query, response_future),
            remote_addr=upstream_addr
        )
        try:
            response = await response_future
        finally:
            transport.close()
        return response

    async def _resolve_by_https(self, query, upstream):
        try:
            with await self.pool.connect(
                upstream.host, 
                upstream.port, 
                Scheme.tls, 
                self._get_proxy(upstream)
            ) as conn:
                q = base64.b64encode(query).decode().rstrip('=')
                return (await https.Client(upstream.hostname, conn).get(
                    upstream.path,
                    {'dns': q}, 
                    [("Content-type", "application/dns-message")]
                )).body
        except HttpException as exc:
            logger.warning('HTTP exception from %s: %s', upstream.name, exc)
        except ConnectionError as exc:
            logger.warning('Failed to connect to %s: %s', upstream.name, exc)

    async def _resolve_by_tls(self, query, upstream):
        try:
            with await self.pool.connect(
                upstream.host, 
                upstream.port, 
                Scheme.tls, 
                self._get_proxy(upstream)
            ) as conn:
                query_size = struct.pack('!H', len(query))
                conn.writer.write(query_size + query)
                await conn.writer.drain()
                response_head = await conn.reader.readexactly(2)
                response_size = struct.unpack('!H', response_head)[0]
                return await conn.reader.readexactly(response_size)
        except asyncio.exceptions.IncompleteReadError:
            logger.error('Failed to read data from %s:%d', upstream.host, upstream.port)
        except ConnectionError as exc:
            logger.warning('Failed to connect to %s: %s', upstream.name, exc)

    async def fetch(self, url, **kwargs):
        return await https.fetch(url, self.handle, 
                                      self.pool, 
                                      self.proxy,
                                      **kwargs)

    def _get_timeout(self, upstream):
        return self.timeout if upstream.timeout is None else upstream.timeout

    def _determine_upstreams(self, kwargs):
        if 'upstreams' in kwargs:
            return kwargs['upstreams']
        elif 'upstream_name' in kwargs:
            upstream_name = kwargs['upstream_name']
            if upstream_name in self.upstreams:
                return list(self.upstreams[upstream_name])
            logger.error('Upstream name %s not defined', upstream_name)
        elif 'upstream_group' in kwargs:
            upstream_group = kwargs['upstream_group']
            if self.upstreams.has_group(upstream_group):
                return self.upstreams.group(upstream_group)
            logger.error('Upstream group %s not defined', upstream_name)
        return self.upstreams.sorted()

    async def handle(self, request, **kwargs):
        logger.debug('DNS query:\n%s', request)
        data = DNSRecord.pack(request)
        
        for upstream in self._determine_upstreams(kwargs):
            if upstream.disable: continue
            if isinstance(upstream, DnsUpstream):
                resolver = self._resolve_by_dns
            elif isinstance(upstream, HttpsUpstream):
                resolver = self._resolve_by_https
            elif isinstance(upstream, TlsUpstream):
                resolver = self._resolve_by_tls
            else:
                raise NotImplementedError('Unspported upstream')

            try:
                with self.stats.record(upstream):
                    response_data = await asyncio.wait_for(
                        asyncio.shield(resolver(data, upstream)), 
                        self._get_timeout(upstream)
                    )
                    if response_data is None:
                        raise EmptyValueError('Empty response data received')
                    try:
                        response = DNSRecord.parse(response_data)
                    except DNSError:
                        raise UnexpectedValueError('Invalid response data received')
                    if request.header.id != response.header.id:
                        raise UnexpectedValueError('Response id does not match')
                    logger.debug('DNS response:\n%s', response)
                    return response
            except ValueError:
                logger.warning('Failed to resolve by upstream server %s', upstream.name)
            except TimeoutError:
                logger.info('Upstream server %s response timeout', upstream.name)

    async def _handle(self, request):
        resolver = self.middlewares if self.middlewares else self
        if request.header.q > 1:
            coroutines = []
            for q in request.questions:
                _req = request.truncate()
                _req.add_question(q)
                coroutines.append(resolver.handle(_req))
            response = DNSRecord(DNSHeader(id=request.header.id,
                                           bitmap=request.header.bitmap,
                                           qr=1, ra=1, aa=1),
                                questions=request.questions)
            for _resp in await asyncio.gather(*coroutines):
                if _resp and _resp.header.a > 0:
                    response.add_answer(_resp.a)
            return response
        return await resolver.handle(request)

    def on_response(self, request, addr, future):
        response = future.result()
        if response is None:
            logger.info('Failed to resolve domain name "%s", upstream servers are unreachable', request.q.qname)
            return

        logger.debug('Send response to %s:%d\n%s' % (addr + (response, )))
        try:
            self.transport.sendto(DNSRecord.pack(response), addr)
        except Exception:
            logger.warning('Failed to send data to %s:%d' % addr)

    def on_request(self, data, addr):
        request = DNSRecord.parse(data)
        logger.debug('Received request from %s:%d\n%s' % (addr + (request, )))
        task = self.loop.create_task(self._handle(request))
        task.add_done_callback(functools.partial(self.on_response, request, addr))

    def abort(self):
        self.status = self.Status.stop_pedding
        logger.info('DNS serivce aborted')
        return self.abort_event.set()

    async def reload(self):
        # TODO: 
        pass

    async def restart(self, silent=False):
        self.status = self.Status.restart_pedding
        if not silent: logger.info('Restarting service')
        self.transport, _ = await self.loop.create_datagram_endpoint(
            lambda: ServerProtocol(self),
            local_addr=(self.host, self.port)
        )
        self.status = self.Status.running
        if not silent: logger.info('DNS service restarted')

    def on_error_restart(self):
        if self.status != self.Status.running:
            return
        return self.loop.create_task(self.restart(True))

    async def run(self):
        self.status = self.Status.start_pedding
        if not await self.middlewares.bootstrap():
            logger.error('Failed to bootstrap')
            return
        try:
            self.transport, _ = await self.loop.create_datagram_endpoint(
                lambda: ServerProtocol(self),
                local_addr=(self.host, self.port)
            )
        except OSError as exc:
            logger.error('Failed to start DNS service: %s', exc)
            return

        self.status = self.Status.running
        logger.info('DNS serivce started')

        try:
            await self.abort_event.wait()
        except asyncio.CancelledError:
            logger.debug('DNS serivce interrupted')
        finally:
            self.transport.close()
            self.status = self.Status.stopped
            logger.info('DNS serivce stopped')

    def open_geoip(self):
        geoip_db = self.config.get('geoip')
        if geoip_db is not None:
            return maxminddb.open_database(geoip_db)

        with resources.open_binary(__package__, 'geoip') as geoip_db:
            return maxminddb.open_database(geoip_db, maxminddb.MODE_FD)



import functools

from urllib.parse import urlsplit
from ipaddress import ip_address


DEFAULT_DOT_PORT = 853

DEFAULT_DNS_PORT = 53

DEFAULT_HTTPS_PORT = 443


class UpstreamCollection:
    def __init__(self, upstreams, only_secure):
        if only_secure:
            upstreams = list(filter(
                lambda up: isinstance(up, (TlsUpstream, HttpsUpstream)), 
                upstreams))
        self._upstreams = upstreams
        self._grouped = dict()
        for upstream in upstreams:
            for group in upstream.groups:
                if group in self._grouped:
                    self._grouped[group].append(upstream)
                else:
                    self._grouped[group] = [upstream]
        self._named = {upstream.name: upstream for upstream in upstreams}
        self._sorted = None
        self._selected = None
        self.sort()

    def _cmp(self, up1, up2):
        if up1.name == self._selected: return 1
        if up2.name == self._selected: return -1
        return up1.priority - up2.priority

    def sort(self):
        self._sorted = self.all()
        self._sorted.sort(reverse=True, key=functools.cmp_to_key(self._cmp))
        for group_name, group_upstreams in self._grouped.items():
            group_upstreams = group_upstreams.copy()
            group_upstreams.sort(reverse=True, key=functools.cmp_to_key(self._cmp))
            self._grouped[group_name] = group_upstreams
        return self

    def all(self):
        return self._upstreams.copy()

    def group(self, name):
        return self._grouped[name]

    def has_group(self, name):
        return name in self._grouped

    def sorted(self):
        if self._sorted is None:
            self.sort()
        return self._sorted

    def select(self, name):
        self._selected = name
        return self.sort()

    @property
    def primary(self):
        return self._sorted[0]

    def __getitem__ (self, name):
        return self._named[name]

    def __contains__(self, name):
        return name in self._named


class Upstream:
    def __init__(self, **kwargs):
        self.name = kwargs.get('name')
        if not isinstance(self.name, str) or self.name == '':
            raise ValueError('Upstream name must be a string')
        self.proxy = kwargs.get('proxy')
        self.timeout = kwargs.get('timeout')
        self.groups = kwargs.get('groups', [])
        if not isinstance(self.groups, list):
            raise ValueError('Upstream groups must be a list')
        self.priority = kwargs.get('priority', 0)
        self.success = 0
        self.usage = 0
        self.disable = False

    def __repr__(self):
        return str(vars(self))

    def to_addr(self):
        return (self.host, self.port)

    def health(self, threshold):
        pass


class DnsUpstream(Upstream):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.host = kwargs['host']
        self.port = kwargs.get('port', DEFAULT_DNS_PORT)


class HttpsUpstream(Upstream):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.url = kwargs['url']
        parsed_url = urlsplit(self.url)
        self.hostname = parsed_url.hostname
        try:
            ip_address(parsed_url.hostname)
            self.host = parsed_url.hostname
        except ValueError:
            self.host = None
        self.port = parsed_url.port if parsed_url.port else DEFAULT_HTTPS_PORT
        self.path = parsed_url.path


class TlsUpstream(Upstream):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.hostname = kwargs['host']
        try:
            ip_address(self.hostname)
            self.host = self.hostname
        except ValueError:
            self.host = None
        self.port = kwargs.get('port', DEFAULT_DOT_PORT)


def parse_upstream(server):
    if isinstance(server, dict):
        server_type = server.get('type', 'https' if 'url' in server else 'dns')
        if server_type == 'dns':
            upstream_class = DnsUpstream
        elif server_type == 'https':
            upstream_class = HttpsUpstream
        elif server_type == 'tls':
            upstream_class = TlsUpstream
        else:
            raise ValueError('Invalid upstream type')
        try:
            return upstream_class(**server)
        except KeyError as e:
            raise ValueError('Missing config key "%s" in "%s"' % (e.args[0], server))
    elif not isinstance(server, str):
        raise TypeError('Parameter server must be dict or string')

    parsed_url = urlsplit(
        server if server.startswith('https://') \
            else '//' + server
    )

    if parsed_url.scheme == 'https':
        return HttpsUpstream(url=server)

    if parsed_url.path == '' and \
       parsed_url.query == '' and \
       parsed_url.fragment == '':
        if parsed_url.port == DEFAULT_DOT_PORT:
            return TlsUpstream(host=parsed_url.hostname, port=DEFAULT_DOT_PORT)
        else:
            return DnsUpstream(host=parsed_url.hostname, 
                               port=DEFAULT_DNS_PORT \
                                   if parsed_url.port is None \
                                   else parsed_url.port)

    raise ValueError('Invalid upstream format "%s"' % (server, ))
